/*
 * secrets.c - Secrets Turbo application
 *
 * Copyright (C) 2003 BLADOX, s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <config.h>
#include <turbo/turbo.h>

#include <stdlib.h>
#include <string.h>

/* *INDENT-OFF* */

lc_char PROGMEM lc_Secrets[]={ 
	LC_EN("Secrets")
	LC_CZ("Tajemstvi")
	LC_END
};

lc_char PROGMEM lc_Name[]={ 
	LC_EN("Name:")
	LC_CZ("Oznaceni:")
	LC_END 
};

lc_char PROGMEM lc_Text[]={ 
	LC_EN("Enter Text:")
	LC_CZ("Vlozit text:")
	LC_END
};

lc_char PROGMEM lc_New_Key[]={ 
	LC_EN("New key:")
	LC_CZ("Novy klic:")
	LC_END
};


lc_char PROGMEM lc_New_Secret[]={
	LC_EN("New Secret")
	LC_CZ("Nove tajemstvi")
	LC_END
};

lc_char PROGMEM lc_Unlock_Key[]={
	LC_EN("Unlock Key")
	LC_CZ("Klic")
	LC_END
};

/* *INDENT-ON* */

u8 PROGMEM def_main_key[] = { 'a', 'd', 'g', 'j', 'm', 'p', 't', 'w' };

typedef struct _Pers_mem
{
  SEdge *secrets;		// pointer to eeprom to key list
  u8 main_key_rnd[16];
  u8 main_key_crnd[16];
  u8 key[16];
}
Pers_mem;

typedef struct _Glob_mem
{
  u8 key[16];
}
Glob_mem;

Twofish_key *key;
Glob_mem *glob_mem;

void make_crypt_key (u8 * buf, u8 * res, u8 key_len)
{
  u8 i;
  u8 tmp[16];

  dbsp ("MAKE_CRYPT_KEY ");
  dbch (key_len);
  dbc ('\n');

  bzero(tmp, 16);
  if (key_len > 16) key_len = 16;
  memcpy(tmp, res, key_len);

  buf[0] = key_len;
  key_len = dcs_78 (tmp, key_len, DCS_8_TO_7);
  for (i = 1; i < 16; i++)
    buf[i] = 16 - key_len;
  for (i = 0; i < key_len; i++)
    buf[i + 1] = tmp[i];
}

u8 secrets_main (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    Pers_mem *p = app_data ();

    ctx->eE = rw (&p->secrets);
  }
  else if (action == APP_LEAVE)
  {
    ctx->eE = NULL;
  }
  return APP_OK;
}

u8 secrets_key (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 *buf = buf_B ();
    u8 *k = buf;
    u8 *in = k + 16;
    u8 *out = in + 16;
    Pers_mem *p = app_data ();
    u8 i;

    i = password (k, 16, locale (lc_New_Key));
    if (i != APP_OK)
      return i;

    twofish_setup (k, 16, key);

    // encrypt secret key with new key
    twofish_ecb_encrypt (&glob_mem->key, &p->key, key);

    // unlock salt with new key
    twofish_ecb_encrypt (&p->main_key_rnd, &p->main_key_crnd, key);

    // reset new key
    twofish_setup (&glob_mem->key, 16, key);

    return APP_BACK;
  }
  return APP_OK;
}

SNodeP secrets_n_view;
SNodeP secrets_n_edit;
SNodeP secrets_n_delete;

u8 secrets_ctx (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    spider_append_r (ctx, &secrets_n_delete);
    spider_append_r (ctx, &secrets_n_edit);
    spider_append_r (ctx, &secrets_n_view);
  }
  else if (action == APP_LEAVE)
  {
    spider_clear_r (ctx);
  }

  return APP_OK;
}

extern SNodeP secrets_n;

u8 secrets_new (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 *res;
    u8 i, j, k;
    SEdge *e = NULL;
    SNode *n = NULL;
    u8 *n_text = NULL;
    u8 *text = NULL;
    Pers_mem *p = app_data ();
    u8 *buf = buf_B ();
    u8 *tmp = buf + 16;

    n = emalloc (sizeof (SNode));
    if (n == NULL)
      goto X_ERR;

    e = emalloc (sizeof (SEdge));
    if (e == NULL)
      goto X_ERR;

    res = get_input (locale (lc_Name), 1, 15, NULL, Q_GET_INPUT_ALPHABET);
    if (res == ENULL)
    {
      i = APP_END;
      goto X_END;
    }
    if (res == NULL)
    {
      i = APP_BACK;
      goto X_END;
    }

    n_text = emalloc (res[0]);
    if (n_text == NULL)
      goto X_ERR;
    memcpy (n_text, res + 2, res[0] - 1);
    wb (n_text + res[0] - 1, '\0');

    res = get_input (locale (lc_Text), 0, 0xE0, NULL, Q_GET_INPUT_ALPHABET);
    if (res == ENULL)
    {
      i = APP_END;
      goto X_END;
    }
    if (res == NULL)
    {
      i = APP_BACK;
      goto X_END;
    }

    // encrypt
    i = res[0] + 8;
    j = i / 16;
    if (i % 16)
      j++;
    text = emalloc (j * 16);
    if (text == NULL)
      goto X_ERR;

    xrand (tmp);
    tmp[8] = j;
    memcpy (tmp + 9, res + 2, res[0] - 1);
    tmp[9 + res[0] - 1] = '\0';

    for (k = 0; k < j; k++)
    {
      if (k)
	for (i = 0; i < 16; i++)
	  tmp[k * 16 + i] ^= buf[i];
      twofish_ecb_encrypt (tmp + k * 16, buf, key);

      memcpy (text + k * 16, buf, 16);
    }

    ww (&n->p, text);
    ww (&n->text, n_text);
    ww (&n->cb, secrets_ctx);

    ww (&e->f, &secrets_n);
    ww (&e->t, n);
    wb (&e->source, MEM_F_P | MEM_T_E);

    ww (&e->next, rw (&p->secrets));
    ww (&p->secrets, e);

    return APP_BACK;
  X_ERR:
    i = APP_OK;
  X_END:
    efree (e);
    efree (n);
    efree (n_text);
    efree (text);

    if (i == APP_OK)
      perror (ERR_NO_EEPROM);
    return i;
  }

  return APP_OK;
}

u8 secrets_view (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    SNode *n = ctx->parent->f;
    u8 j = 1;
    u8 i, k;
    u8 *buf = buf_B ();
    u8 *in = buf + 16;
    u8 *tmp = in + 16;
    u8 *t = rw (&n->p);

    i = 0;
    while (i < j && i < 16)
    {
      memcpy (tmp + i * 16, t + i * 16, 16);

      twofish_ecb_decrypt (tmp + i * 16, buf, key);

      if (i)
	for (k = 0; k < 16; k++)
	  buf[k] = buf[k] ^ in[k];

      memcpy (in, tmp + i * 16, 16);
      memcpy (tmp + i * 16, buf, 16);

      if (i == 0)
	j = tmp[8];

      i++;
    }

    return display_text (tmp + 9, NULL);
  }

  return APP_OK;
}

u8 secrets_edit (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    SNode *n = ctx->parent->f;
    u8 j = 1;
    u8 i, k;
    u8 *buf = buf_B ();
    u8 *in = buf + 16;
    u8 *tmp = in + 16;
    u8 *t = rw (&n->p);
    u8 *res;
    u8 *s;

    i = 0;
    while (i < j && i < 16)
    {
      memcpy (tmp + i * 16, t + i * 16, 16);
      twofish_ecb_decrypt (tmp + i * 16, buf, key);

      if (i)
	for (k = 0; k < 16; k++)
	  buf[k] = buf[k] ^ in[k];

      memcpy (in, tmp + i * 16, 16);
      memcpy (tmp + i * 16, buf, 16);

      if (i == 0)
      {
	j = tmp[8];
      }

      i++;
    }

    res = get_input (locale (lc_Text), 0, 0xE0, tmp + 9, Q_GET_INPUT_ALPHABET);
    if (res == ENULL)
      return APP_END;
    if (res == NULL)
      return APP_BACK;
    i = res[0] + 8;
    j = i / 16;
    if (i % 16)
      j++;

    s = emalloc (j * 16);
    if (s == NULL)
      return perror (ERR_NO_EEPROM);
    efree (t);
    t = s;

    xrand (tmp);
    tmp[8] = j;
    memcpy (tmp + 9, res + 2, res[0] - 1);
    tmp[9 + res[0] - 1] = '\0';

    for (k = 0; k < j; k++)
    {
      if (k)
	for (i = 0; i < 16; i++)
	  tmp[k * 16 + i] ^= buf[i];

      twofish_ecb_encrypt (tmp + k * 16, buf, key);
      memcpy (t + k * 16, buf, 16);
    }

    ww (&n->p, t);

    return APP_BACK;
  }

  return APP_OK;
}

u8 secrets_delete (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 res;
    SNode *n = ctx->parent->f;

    res = yesno (rw (&n->text), Q_YESNO_DELETE_QMARK, 0);
    if (res == APP_END)
      return APP_END;
    if (res == APP_YES)
    {
      Pers_mem *p = app_data ();
      SEdge *e = rw (&p->secrets);
      SEdge *f = e;

      while (e)
      {
	if (rw (&e->t) == n)
	  break;
	f = e;
	e = rw (&e->next);
      }

      if (f == e)
	ww (&p->secrets, rw (&e->next));
      else
	ww (&f->next, rw (&e->next));

      efree (e);
      efree (rw (&n->text));
      efree (rw (&n->p));
      efree (n);
      spider_back (ctx);
    }
    return APP_BACK;
  }

  return APP_OK;
}

SNodeP secrets_n = { lc_Secrets, secrets_main };
SNodeP secrets_n_key = { lc_Unlock_Key, secrets_key };
SNodeP secrets_n_new = { lc_New_Secret, secrets_new };
SNodeP secrets_n_view = { LC_VIEW, secrets_view };
SNodeP secrets_n_edit = { LC_EDIT, secrets_edit };
SNodeP secrets_n_delete = { LC_DELETE, secrets_delete };

SEdgeP secrets_edges_p[]={
{&secrets_n, &secrets_n_key},
{&secrets_n, &secrets_n_new},
NULL
};

void secrets (void *x)
{
  SCtx *c;
  u8 i;
  u8 *res;
  u8 err = NO_ERROR;
  u8 *buf = buf_B ();
  u8 *k = buf;
  u8 *in = buf + 16;
  u8 *out = in + 16;
  Pers_mem *p = app_data ();

  glob_mem = NULL;

  i = password (k, 16, NULL);
  if (i != APP_OK)
    return;

  key = malloc (sizeof (Twofish_key));
  if (key == NULL)
  {
    perror (ERR_NO_RAM);
    return;
  }

  glob_mem = malloc (sizeof (Glob_mem));
  if (glob_mem == NULL)
  {
    err = ERR_NO_RAM;
    goto X_ERR;
  }

  err = twofish_setup (k, 16, key);
  if (err != NO_ERROR)
    goto X_ERR;

  twofish_ecb_encrypt (&p->main_key_rnd, out, key);

    if (memcmp(out, p->main_key_crnd, 16))
    {
      err = ERR_WRONG_KEY;
      goto X_ERR;
    }

  twofish_ecb_decrypt (&p->key, &glob_mem->key, key);

#ifdef DEBUG
  dbsp ("SECRET_KEY_1: ");
  for (i = 0; i < 16; i++)
  {
    dbch (out[i]);
    dbc (' ');
  }
  dbc ('\n');
#endif

  err = twofish_setup (&glob_mem->key, 16, key);
  if (err != NO_ERROR)
    goto X_ERR;

  c = spider_init ();
  if (c != NULL)
  {
    c->n = &secrets_n;
    c->eP = &secrets_edges_p;
    c->n_default = 2;

    spider (c);
  }
  else
    err = ERR_NO_RAM;

X_ERR:
  if (err != NO_ERROR)
    perror (err);
  free (key);
  key = NULL;
  free (glob_mem);
  glob_mem = NULL;
}

void turbo_handler (u8 action, void *data)
{
  switch (action)
  {
    case ACTION_APP_REGISTER:
      {
	Pers_mem *p = emalloc (sizeof (Pers_mem));
	Twofish_key *ak = malloc (sizeof (Twofish_key));
	u8 i;
	u8 *buf = malloc (2 * 16);
	u8 *in = buf;
	u8 *tmp = in + 16;

	if (p && buf)
	{
	  ww (&p->secrets, NULL);
	  memcpy (tmp, def_main_key, sizeof (def_main_key));
	  make_crypt_key (in, tmp, sizeof (def_main_key));
#ifdef DEBUG
	  dbsp ("DEF_KEY: ");
	  for (i = 0; i < 16; i++)
	  {
	    dbch (in[i]);
	    dbc (' ');
	  }
	  dbc ('\n');
#endif
	  twofish_setup (in, 16, ak);

	  xrand (in);
	  xrand (in + 8);
	  memcpy (&p->main_key_rnd, in, 16);	//salt
	  twofish_ecb_encrypt (in, &p->main_key_crnd, ak); //crypted salt
	  
	  xrand (in);
	  xrand (in + 8);
#ifdef DEBUG
	  dbsp ("SECRET_KEY: ");
	  for (i = 0; i < 16; i++)
	  {
	    dbch (in[i]);
	    dbc (' ');
	  }
	  dbc ('\n');
#endif

	  twofish_ecb_encrypt (in, &p->key, ak); // random key used for content encryption
#ifdef DEBUG
	  dbsp ("KEY_RND: ");
	  for (i = 0; i < 16; i++)
	  {
	    dbch (rb (&p->main_key_rnd[i]));
	    dbc (' ');
	  }
	  dbc ('\n');
	  dbsp ("KEY_CRND: ");
	  for (i = 0; i < 16; i++)
	  {
	    dbch (rb (&p->main_key_crnd[i]));
	    dbc (' ');
	  }
	  dbc ('\n');

#endif
	}
	reg_app_data (p);

	free (buf);
	free (ak);
      }
      break;
    case ACTION_APP_UNREGISTER:
      {
	Pers_mem *p = app_data ();
	SEdge *e, *f;
	SNode *n;
	u8 *t;

	e = rw (&p->secrets);
	while (e)
	{
	  n = rw (&e->t);
	  t = rw (&n->text);
	  efree (rw (&n->p));
	  efree (t);
	  efree (n);
	  f = rw (&e->next);
	  efree (e);
	  e = f;
	}
	efree (p);
      }
      break;
    case ACTION_INSERT_MENU:
      insert_menu (locale (lc_Secrets));
      break;
    case ACTION_MENU_SELECTION:
      stk_thread (secrets, NULL);
      break;
    default:
      break;
  }
}
